#!/usr/bin/perl

#This program will scan a genome to find duplicate genes and calculate Ks and 
#Ka for those genes.  Please see the included documentation for details
#Copyright 2002 Gavin Conant and Andreas Wagner



use strict 'vars';

my($error, $paramfile, $protein_seq_file, $nucleotide_seq_file, $pairs_file, @lines, $i, $len,
   $results_file, $partners, $genename, $pro_line, $clustout, $tempseqfile, @pro_part, @seq1, @seq2, @seq3, @seq4,
   $error_file, $percent_sim, $nucseq_file1, $nucseq_file2, $lenseq1, $lenseq2, $lenseq3, $lenseq4, $tempparam, $nucalignfile,
   $align_check, $likeout, $ks, $ka,  $BLAST_matloc, $top_BLAST, $BLAST_thresh, $percent_thresh, $genename_field,
   @genelist, @degreelist, @proseqs, @nucseqs, $numgenes, $j, $minorflen, $min_algn_res, $num_aln_res, $skip_match, 
   $name1, $name2, $cmdargs, $percent_sim2, %completed_pairs, $pair_name, $seq_num, $header, $checkpoint_file, $checkpoint_on, $start_run,
   $old_out, $old_err, @out_lines, @err_lines, $check_name, $len, $date, $sat, $align_error,
   $current_checkpoint, $num_aln_res2, $gencode_arg, $gen_code, $warnings, $accept, $ordered_genes, $new_protein_seq_file, $basefreqs, $basefreq_arg);

$cmdargs=@ARGV;

#Stores the list of analyzed gene pairs
%completed_pairs = ();

if (($cmdargs > 1) || ($ARGV[0] =~ /\w{1,2}=/)) {
    print "Using command-line arguments\n";
    @lines = @ARGV; 
    $paramfile = "\0";
}
elsif ($cmdargs == 1) {
    $paramfile=$ARGV[0];
    print "Reading options from file $paramfile\n";
    open (READPARAM, "<" . $paramfile) or die "Can't open input parameter file $paramfile\n";
    @lines = <READPARAM>;
}

($protein_seq_file, $nucleotide_seq_file, $pairs_file, $results_file,
 $error_file, $top_BLAST, $BLAST_matloc, $BLAST_thresh, $percent_thresh, $minorflen, $min_algn_res, $genename_field,$gencode_arg, $gen_code, $warnings, $ordered_genes, $basefreqs, $checkpoint_on) =read_param_file(@lines);


if ($paramfile ne "\0") {
    $checkpoint_file = ".GenomeHistory.chkpnt." . $paramfile;
}
else {
    $checkpoint_file = ".GenomeHistory.chkpnt.cmdline";
}



#If no arguments were given, list them
if ($cmdargs == 0) {
    print "Usage: GenomeHistory.pl <command file> or \n\tOption\t\t\t\t\tDefault\n\t---------------------------------------------------------\n";
    print "\tpi=<protein sequences file>\t\t$protein_seq_file\n\tni=<nucleotide_sequences_file>\t\t$nucleotide_seq_file\n\to=<output file>\t\t\t\t$results_file\n\te=<error file>\t\t\t\t$error_file\n\tgf=<Gene name file>\t\t\t$pairs_file\n\tgn=<Gene name field #>\t\t\t$genename_field\n\tgc=<U/VM/YM/IM/MM/CN/EM>\t\t$gen_code\n\tbl=<BLAST BLOSUM matrix location>\tNA\n\ttb=<YES/NO>\t\t\t\t$top_BLAST\n\tbt=<BLAST E-value threshold>\t\t$BLAST_thresh\n\tmo=<Minium Sequence length>\t\t$minorflen\n\tma=<Minimum # aligned residues>\t\t$min_algn_res\n\tpt=<Minimum percent identity>\t\t$percent_thresh\n\tw=<ON/OFF>\t\t\t\t$warnings\n\tog=<TRUE/FALSE>\t\t\t\t$ordered_genes\n\tbf=<OBSERVED/CODON/EQUAL>\t\t$basefreqs\n\tcp=<ON/OFF>\t\t\t\t$checkpoint_on\n";
	exit();
}

if (($checkpoint_on =~ /on/)) {
    print "Checkpointing is ON: Checkpoint file $checkpoint_file";
    if (-e $checkpoint_file) {
	print " exists\n";
    }
    else {
	print" does not exist\n";}
}
else {
    print "Checkpointing is OFF\n";  
}
if (($checkpoint_on =~ /on/) && (-e $checkpoint_file)) {
	open (READCHKPNT, "<" . $checkpoint_file) or die "Can't open checkpoint file $checkpoint_file\n";
	$current_checkpoint = <READCHKPNT>;
	$current_checkpoint =~ s/\n//;
	close (READCHKPNT);



	print "Restarting from checkpoint: Gene: $current_checkpoint\n";

	open(INPUTKSKA, "<" . $results_file) or die "Can't open output file $results_file\n";
	open(INPUTERRORS, "<" . $error_file) or die "Can't open error output file $error_file\n";
	@out_lines=<INPUTKSKA>;
	@err_lines=<INPUTERRORS>;
	close(INPUTKSKA);
	close(INTPUTERRRORS);
	
	$j=0;


	if (!(genename_field =~ /none/i))
	{
	    $check_name=get_gene_name($current_checkpoint, $genename_field);
	}
	else 
	{
	    $check_name=$current_checkpoint;
	}
	
	
	open(OUTPUTKSKA, ">" . $results_file) or die "Can't open output file $results_file\n";
	open(OUTPUTERRORS, ">" . $error_file) or die "Can't open error output file $error_file\n";
	
	$len=@out_lines;
	$j=0;
	

	while (!($out_lines[$j] =~ /^$check_name/) && ($j < $len)) 
	{ 
	    print OUTPUTKSKA "$out_lines[$j]";
	    $j++;
	}
	
	$len=@err_lines;
	$j=0;
	while (!($err_lines[$j] =~ /^$check_name/) && ($j < $len)) 
	{ 
	    print OUTPUTERRORS "$err_lines[$j]";
	    $j++;
	}
	

    }
else {
    $current_checkpoint = "\0";
#Checks before overwriting output files
    if (-e $results_file) {
	print "Error: results file $results_file exists: Overwrite? (YES/NO) ";
	$accept=<STDIN>;
	if (!($accept =~ /yes/i)) {
	    print "Aborting\n";
	    exit();
	}   
    }
    
    
    if (-e $error_file) {
	print "Error: error file $error_file exists: Overwrite? (YES/NO) ";
	$accept=<STDIN>;
	if (!($accept =~ /yes/i)) {
		print "Aborting\n";
		exit();
	    }
    }
    
    open(OUTPUTKSKA, ">" . $results_file) or die "Can't open output file $results_file\n";
    open(OUTPUTERRORS, ">" . $error_file) or die "Can't open error output file $error_file\n";
    
    
}


#Stupid test to see if we can write checkpoint file
open (WRITECHKPNT, ">" . $checkpoint_file) or die "Can't open checkpoint file $checkpoint_file\n";
close (WRITECHKPNT);

open(INPUTPROTEINSEQ, "<" . $protein_seq_file) or die "Can't open protein sequence file $protein_seq_file\n";
open(INPUTNUCSEQ, "<" . $nucleotide_seq_file) or die "Can't open nucleotide sequence file $nucleotide_seq_file\n";
@proseqs = <INPUTPROTEINSEQ>;
@nucseqs = <INPUTNUCSEQ>;
close(INPUTPROTEINSEQ);
close(INPUTNUCSEQ);






#Formats Protein sequence file to avoid BLAST problems with | characters:
$new_protein_seq_file = $protein_seq_file . ".tmp";
print "Formating protein sequences...Please wait";
open (OUTPUT_TEMP_PROSEQ, ">" . $new_protein_seq_file) or die "Can't open temporary protein sequence file $new_protein_seq_file\n";
$len=@proseqs;

for ($i=0; $i<$len; $i++) {
    if ($proseqs[$i] =~ /^>/) { 
	if ($proseqs[$i] =~ /\|/) {
	    $proseqs[$i] =~ s/://g;
	    $proseqs[$i] =~ s/\|/:/g;
	}
    }
    print OUTPUT_TEMP_PROSEQ "$proseqs[$i]";
}
close (OUTPUT_TEMP_PROSEQ);
$protein_seq_file = $new_protein_seq_file;
open(INPUTPROTEINSEQ, "<" . $protein_seq_file) or die "Can't open protein sequence file $protein_seq_file\n";

@proseqs=0;
@proseqs = <INPUTPROTEINSEQ>;
close(INPUTPROTEINSEQ);
print "\nDone\n";




#Print current configuration
if ($current_checkpoint eq "\0"){
    $date = `date`;
    print OUTPUTKSKA "Runtime: $date\n";    
    print OUTPUTKSKA "Running GenomeHistory.pl.   Settings are:\n Protein sequences file: $protein_seq_file\n ORFs file: $nucleotide_seq_file\n Field where gene names are found: $genename_field\n";
    if (!($pairs_file =~ /NONE/)) {
	print OUTPUTKSKA " Gene list file: $pairs_file\n";
    }

    

    print OUTPUTKSKA " Output file: $results_file\n Error file: $error_file\n Minimum translated ORF length: $minorflen\n Minimum number of aligned residues to accept pair: $min_algn_res\n";
    
    if (!($BLAST_matloc =~ /NONE/)) {
	print OUTPUTKSKA " Using BLAST matrix $BLAST_matloc\n";
    }
    
    if ($top_BLAST =~ /yes/i) {
	print OUTPUTKSKA " Only accepting top BLAST hit for each gene\n";
    }
    else {
	print OUTPUTKSKA " Accepting all BLAST hits with e <= $BLAST_thresh\n";
    }
    if ($ordered_genes =~ /TRUE/i) {
	print OUTPUTKSKA " Nucleotide sequences identified by sequence file order\n";
    }
    
    print OUTPUTKSKA " Base Frequencies used: $basefreqs\n";
    print OUTPUTKSKA " Minimum percent identity for analysis: $percent_thresh\n"; 
    print OUTPUTKSKA " Genetic code: $gen_code\n Ks values marked with * are likely saturated\n\n\n";
    print OUTPUTKSKA "Gene1\tGene2\tKs\tKa\n";
}


#Create the BLAST database from input protein sequences
$error = `nice setdb $protein_seq_file`;

if ($error != 0) 
{
    die "ERROR: failed to create BLAST database from protein sequence file $protein_seq_file.\nThe error reported was $error\n";
}


#We will either get the list of genes to BLAST from the protein sequence file
if ($pairs_file =~ /NONE/) {
    @genelist=get_full_genelist($protein_seq_file, $genename_field);   
}
#Or from a file given by the user
else {
    @genelist=get_file_genelist($pairs_file);
    @degreelist=get_gene_partners($pairs_file, @genelist);
}

$numgenes=@genelist;

$nucseq_file1=  "Gene1.fas";
$nucseq_file2= "Gene2.fas";
$nucalignfile = "nuc_temp.nex";
												
if ($basefreqs =~ /OBSERVED/i) { $basefreq_arg="-obs_freqs"; }
else { 
    if ($basefreqs =~ /EQUAL/i) { $basefreq_arg=""; }
    else { $basefreq_arg="-codon_freqs"; } 
}




#If we have a checkpoint, start from there
if (($checkpoint_on =~ /on/) && ($current_checkpoint ne "\0")) {
    $j=0;
    while (!($genelist[$j] =~ /$current_checkpoint/) && ($j<$numgenes)) {$j++;}
    $start_run = $j;
}
else {$start_run=0;}



for($j=$start_run; $j<$numgenes; $j++)
{
    if (!($pairs_file =~ /NONE/)) {
	$partners = $degreelist[$j];
    }
    $genename=0;
    $genename=$genelist[$j];

    open (WRITECHKPNT, ">" . $checkpoint_file) or die "Can't open checkpoint file $checkpoint_file\n";
    print WRITECHKPNT "$genename\n";
    close (WRITECHKPNT);

    $seq1[0]="NONE";
    ($header, @seq1)=find_sequence($genename, "", @proseqs);    
    $lenseq1=get_seqlen(@seq1);
   
    if (!($seq1[0] =~ /NONE/)) {
	if ($lenseq1 >= $minorflen) {
	    $tempseqfile="tempseq.fas";

	    if (!(genename_field =~ /none/i))
	    {
		$name1=get_gene_name($genename, $genename_field);
	    }
	    else 
	    {
		$name1=$genename;
	    }

	    if (!($pairs_file =~ /NONE/)) {
		$genename=$header;
	    }

	    $seq3[0] = "NONE";
	    write_fasta_file($tempseqfile, $genename, @seq1);
	    @pro_part=0;
	    @pro_part=parse_blastout($genename, $tempseqfile, $top_BLAST, $BLAST_thresh, $BLAST_matloc, $genename_field);
	    
#if we are only taking the top hit, @pro_part will have only 1 entry
	    if (!($pro_part[0] =~ /NONE/)) 
	    {
		$len =@pro_part;
		
	    

		for ($i=0; $i<$len; $i++)
		{
		    $skip_match="FALSE";
		
		    if (!(genename_field =~ /none/i))
		    {
			$name2=get_gene_name($pro_part[$i], $genename_field);
		    }
		    else 
		    {
			$name2=$pro_part[$i];
		    }
		    $pair_name=$name2 . "," . $name1;

    
		    if (($pairs_file =~ /NONE/) && ($top_BLAST =~ /no/i)) {
			if ($completed_pairs{$pair_name} ne "" ) {
			    $skip_match="TRUE";
			}
			else {
			    $pair_name=$name1 . "," . $name2;
			    $completed_pairs{$pair_name} = "DONE";
			}
		    }
		   
		    
		    if ($skip_match =~ /FALSE/) {
			$seq2[0]="NONE";
			($header, @seq2)=find_sequence($pro_part[$i], "",  @proseqs);
			$lenseq2=get_seqlen(@seq2);
			
			if (!($seq2[0] =~ /NONE/)) { 
			    if ($lenseq2 >= $minorflen) 
			    { 
#				print "Writing protein FASTA file. Length of sequence #2 ($name2): $lenseq2\n";
				$tempseqfile="tempseq.fas";
	   
				write_fasta_file($tempseqfile, "Gene1", @seq1, ">Gene2", @seq2);
				$clustout=parse_clustal_output($tempseqfile);		
				if ($clustout == 1) {
				   

				    $tempseqfile =~ s/fas/pir/;
				    ($percent_sim, $num_aln_res)=find_percent_diff($tempseqfile, "protein");
				    print "Percent identity for alignment of $name1 and $name2: $percent_sim, Number of aligned residues: $num_aln_res\n";
				    if ($num_aln_res >= $min_algn_res) {
					if ($percent_sim>$percent_thresh) {
					    if ($seq3[0] =~ /NONE/i) {
						#Only get this sequence once.
						if ($ordered_genes =~ /TRUE/i) {
						    $seq_num=$j;

						    if (!($pairs_file =~ /NONE/)) {
							$seq_num=find_sequence_num($name1, "", @proseqs);
						    }
						    
						    @seq3=find_sequence_w_num($seq_num, @nucseqs);
						}
						else {
						    ($header, @seq3)=find_sequence($genename, $name1, @nucseqs);
						    
						}
						if (!($seq3[0] =~ /NONE/)) {
						    $lenseq3=get_seqlen(@seq3);
						    write_fasta_file($nucseq_file1, "Gene1", @seq3);
						}
					    }

					    if ($ordered_genes =~ /TRUE/i) {
						$seq_num=find_sequence_num($pro_part[$i], "", @proseqs);
						print "Searching for sequence $pro_part[$i]:  Apparent number is $seq_num\n";
						if ($seq_num != -1) {
						    @seq4=find_sequence_w_num($seq_num, @nucseqs);}
						else {
						    $seq4[0] = "NONE";
						}
					    }
					    else {
						($header, @seq4)=find_sequence($pro_part[$i], $name2, @nucseqs);
					    }

					    
					    if (!($seq3[0] =~ /NONE/) && !($seq4[0] =~ /NONE/)) {
												
						$lenseq4=get_seqlen(@seq4);
						
						write_fasta_file($nucseq_file2, "Gene2", @seq4);
						$tempparam="tempparams.txt";
						open(WRITEPARAMFILE, ">" . $tempparam) or die "Can't open temp output file $tempparam\n";
						print WRITEPARAMFILE "Gene1 1 $lenseq3\n";
						print WRITEPARAMFILE "Gene2 1 $lenseq4\n";
						close (WRITEPARAMFILE);
						
						$align_check = nuc_align ($nucalignfile, $tempseqfile, $tempparam, $gencode_arg);
						
						if (!$align_check) {
						    
						    ($percent_sim2, $num_aln_res2)=find_percent_diff($nucalignfile, "dna");
						    if ($percent_sim2 <1 ) {
							$likeout = `nice like_pair_dist $nucalignfile -nogap $basefreq_arg $gencode_arg`;
							
							($ks, $ka, $sat)=find_ks_ka($likeout);
							
							print OUTPUTKSKA "$name1\t$name2\t$ks$sat\t$ka\n";
						    }
						    else
						    {
							print "$name1\t$name2\tIDENTICAL\n";
							print OUTPUTKSKA "$name1\t$name2\tIDENTICAL\n";
							print OUTPUTERRORS "WARNING: $name1 and $name2 have identical nucleotide sequences\n";
						    }
						    
						}
						else {
						    $align_error=nuc_align_error($align_check);
						    print OUTPUTERRORS "ERROR: Nucleotide alignment of $name1 and $name2 failed--Pair skipped\n";
						    print OUTPUTERRORS "$align_error\n";
	}
						if (-e  $nucalignfile) {
						    `rm $nucalignfile`;}
						if (-e $nucseq_file2){
						    `rm $nucseq_file2`;}
					    }
					    else {
						print OUTPUTERRORS "ERROR: One or both nucleotide sequences missing for pair $name1, $name2--Pair skipped\n"; 
					    }
					}
					else {
					    if ($warnings =~ /on/i) {
					    print OUTPUTERRORS "WARNING: Sequences $name1 and $name2 have identity of $percent_sim:  They will be excluded from the analysis\n";} 
					}
				    }
				    else {
					if ($warnings =~ /on/i) {
					print OUTPUTERRORS "WARNING: Sequences $name1 and $name2 have only $num_aln_res aligned residues:  They will be excluded from the analysis\n";}  
				    }
				    
				}
				else {
				    print OUTPUTERRORS "ERROR: Clustal alignment of sequences $name1 and $name2 failed\n";
				}
			    }
			    else {
				if ($warnings =~ /on/i) {
				    print OUTPUTERRORS "WARNING: Protein sequence for gene $name2 (partner of $name1) below minimum length--Omitting\n";}
			    }
			}
			else {
			    print OUTPUTERRORS "ERROR: Protein sequence for gene $name2 (partner of $name1) not found\n";
			}
		    }
		}
	    }
	    else 
	    { 
		if ($warnings =~ /on/i) {
		    print OUTPUTERRORS "WARNING: gene $name1 has no paralogs\n";}
	    }
	}
	else {
	    if ($warnings =~ /on/i) {
	    print OUTPUTERRORS "WARNING: Protein sequence for gene $genename below minimum length--Omitting\n";}  
	}	
    }
    else {
	print OUTPUTERRORS "ERROR: Protein sequence for gene $genename not found\n";
    }
}
close (OUTPUTKSKA);
close (OUTPUTERRORS);

if (-e $nucseq_file1){
    `rm $nucseq_file1`;}
if (-e "tempseq.fas")
{
    `rm tempseq.fas`;
}
if (-e "tempseq.pir")
{
    `rm tempseq.pir`;
} 
if (-e "tempseq.dnd")
{
    `rm tempseq.dnd`;
} 
if (-e $tempparam) {
    `rm $tempparam`;
}
if (-e $new_protein_seq_file) {
    `rm $new_protein_seq_file`;
}
if (-e $checkpoint_file ) {
    `rm $checkpoint_file`;
}


sub read_param_file {
    my (@lines) = @_;
    my( $i, $len, $protein_seq_file, $nucleotide_seq_file, $pairs_file, $results_file,
       $error_file, $top_BLAST, $BLAST_matloc, $BLAST_thresh, $percent_thresh, $minorflen, $min_algn_res, $genename_field, $gencode_arg
	, $gen_code, $warnings, $ordered_genes, $basefreqs, $checkpoint_on);

    $gencode_arg="";
    $gen_code="Universal";
    $warnings="OFF";
    $BLAST_matloc = "NONE";
    $BLAST_thresh = 0.01;
    $percent_thresh = 0.50;
    $top_BLAST="NO";
    $pairs_file = "NONE";
    $minorflen=40;
    $min_algn_res=40;
    $genename_field=2;
    $protein_seq_file = "orf_trans.fasta";
    $nucleotide_seq_file = "orf_coding.fasta";
    $results_file="genome_ks.txt";
    $error_file="genome_ks.err";
    $ordered_genes=FALSE;
    $basefreqs="CODON";
    $checkpoint_on="OFF";

    $len=@lines;

#Read the parameter file
    for ($i=0; $i<$len; $i++)
    {
	if (!($lines[$i] =~ /\A\#.{1,300}/))
	    #Matches non-comment lines
	{
	    if (($lines[$i] =~ /protein\ssequences\s{0,1}=\s{0,1}(.{2,70})/i) || 
		($lines[$i] =~ /^pi\s?=\s?(.{2,70})/i))
	    {
		$protein_seq_file=$1;
	    }
	    if (($lines[$i] =~ /orf\ssequences\s{0,1}=\s{0,1}(.{2,70})/i) || 
		($lines[$i] =~ /^ni\s?=\s?(.{2,70})/i))
	    {
		$nucleotide_seq_file=$1;
	    }
	    if (($lines[$i] =~ /gene\sfile\s{0,1}=\s{0,1}(.{2,70})/i) ||
		($lines[$i] =~ /^gf\s?=\s?(.{2,70})/i))
	    {
		$pairs_file=$1;
	    }
	    if (($lines[$i] =~ /output\sfile\s{0,1}=\s{0,1}(.{2,70})/i) ||
		($lines[$i] =~ /^o\s?=\s?(.{2,70})/i))
	    {
		$results_file=$1;
	    }
	    if (($lines[$i] =~ /error\sfile\s{0,1}=\s{0,1}(.{2,70})/i) ||
		($lines[$i] =~ /^e\s?=\s?(.{2,70})/i))
	    {
		$error_file=$1;
	    }
	    if (($lines[$i] =~ /only\stop\sblast\smatch/i) || ($lines[$i] =~ /^tb\s?=\s?/i))
	    {
		if ($lines[$i] =~ /yes/i) {
		    $top_BLAST="YES";}
	    } 
	    if ($lines[$i] =~ /blast\smatrix\slocation/i)
	    {
		$lines[$i] =~ s/blast\smatrix\slocation\s{0,1}=\s{0,1}//i;
		$lines[$i] =~ s/\n//;
		$BLAST_matloc=$lines[$i];	
	    } 
	    if ($lines[$i] =~ /^bl\s?=\s?/i)
	    {
		$lines[$i] =~ s/bl\s?=\s?//i;
		$lines[$i] =~ s/\n//;
		$BLAST_matloc=$lines[$i];	
	    }
	    if (($lines[$i] =~ /blast\sthreshold\s{0,1}=\s{0,1}(\d{1,12}\.{0,1}\d{0,12}e?-?\d{0,5})/i) || 
		($lines[$i] =~ /^bt\s?=\s?(\d{1,12}\.{0,1}\d{0,12}e?-?\d{0,5})/i))
	    {
		$BLAST_thresh=$1;
		$BLAST_thresh*=1.0;
	    }
	    if (($lines[$i] =~ /genetic\scode\s?=\s?(\w{1,2})/) || ($lines[$i] =~ /gc\s?=\s?(\w{1,2})/)) {
		$gencode_arg= "-v:" . $1;
		$gen_code=name_genetic_code($1);
	    }

	    if (($lines[$i] =~ /percent\sidentity\sthreshold\s{0,1}=\s{0,1}(\d\.\d{0,12})/i) ||
		($lines[$i] =~ /^pt\s?=\s?(\d\.\d{0,12})/i))
	    {
		$percent_thresh=$1;
	    } 
	    if (($lines[$i] =~ /minimum\sorf\stranslation\slength\s{0,1}=\s{0,1}(\d{1,10})/i) || 
		($lines[$i] =~ /^mo\s?=\s?(\d{1,10})/i))
	    {
		$minorflen=$1;
	    } 
	    if (($lines[$i] =~ /minimum\snumber\saligned\sresidues\s{0,1}=\s{0,1}(\d{1,10})/i) ||
		($lines[$i] =~ /^ma\s?=\s?(\d{1,10})/i))
	    {
		$min_algn_res=$1;
	    } 
	    if (($lines[$i] =~ /warnings\s?=\s?(\w{2,3})/i) ||
		($lines[$i] =~ /^mw\s?=\s?(\w{2,3})/i))
	    {
		$warnings=$1;
	    }
	    if (($lines[$i] =~ /gene\sname\sfield\s{0,1}=\s{0,1}(\d{1,10})/i) || 
		($lines[$i] =~ /^gn\s?=\s?(\d{1,10})/i))
	    {
		$genename_field=$1;
	    }
	    if (($lines[$i] =~ /checkpoint\s{0,1}=\s{0,1}(\w{1,3})/i) || ($lines[$i] =~ /^cp\s?=\s?(\w{2,3})/i))
	    {
		if ($1 =~ /on/i) {
		    $checkpoint_on = "on";}
	    }

	    if (($lines[$i] =~ /ordered\sgene\slist\s{0,1}=\s{0,1}/i) || ($lines[$i] =~ /^og\s?=\s?/i)) {
		if ($lines[$i] =~ /yes/i) {
		    $ordered_genes="TRUE";}
	    }
	
	    if (($lines[$i] =~ /base\sfrequencies\s{0,1}=\s{0,1}/i) || ($lines[$i] =~ /^bf\s?=\s?/i)) {
		$lines[$i] =~ s/base\sfrequencies\s{0,1}=\s{0,1}//i;
		$lines[$i] =~ s/^bf\s?=\s?//i;
		$lines[$i] =~ s/\n//;
		$basefreqs = $lines[$i];
	    }
	}
    }
    return($protein_seq_file, $nucleotide_seq_file, $pairs_file, $results_file,
       $error_file, $top_BLAST, $BLAST_matloc, $BLAST_thresh, $percent_thresh, $minorflen, $min_algn_res, $genename_field, $gencode_arg, $gen_code, $warnings, $ordered_genes, $basefreqs, $checkpoint_on);


    }



sub name_genetic_code {
    my ($arg) = @_;
    my ($name);

    if ($arg =~ /^u/i) {
	$name="Universal";
    }
    if ($arg =~ /^vm/i) {
	$name="Vertebrate Mitochondrial";
    }
    if ($arg =~ /^ym/i) {
	$name="Yeast Mitochondrial";
    }
    if ($arg =~ /^mm/i) {
	$name="Mold mitochondrial";
    } 
    if ($arg =~ /^im/i) {
	$name="Invertebrate mitochondrial";
    } 
    if ($arg =~ /^cn/i) {
	$name="Ciliate nuclear";
    }
    if ($arg =~ /^em/i) {
	$name="Echinoderm mitochondrial";
    }
    return($name);
}


sub get_gene_name {
    my($FASTA_name, $gene_field) =@_;
    my ($i, $line, $temp);

    $line=$FASTA_name;
    $i=0;
   
# Clean-up lousy formating in FASTA records
    $line=clean_line($line);
   
    while (($i < $gene_field) && ($line ne "")) {
	$temp="";
	$line =~ /([^:&^\n&^\#]{2,2000}(:|\n|\#))/;
	$temp=$1;
	$line =~ s/$temp//;
	$line =~ s/^(:|\n|\#)//;
	$i++;
    }
    if ($temp eq "") {
	$temp=$line;}

    $temp =~ s/\n//g;
    $temp =~ s/://g;
    $temp =~ s/\#//g;
    $temp =~ s/^\s//;
    
    $temp =~ /[^\s]{1,100}(\s{1,5}.{1,2000})/;
    $i=$1;
    $temp =~ s/$i//;
   
    return($temp);
}


sub clean_line {
    my ($line) =@_;
    $line =~ s/>//;
    $line =~ s/\n//; 
    $line =~ s/\|/:/g;
    $line =~ s/\\//g;
    $line =~ s/\///g;
    $line =~ s/\(/%/g;
    $line =~ s/\)/%/g;
    $line =~ s/\[/%/g;
    $line =~ s/\]/%/g;
    $line =~ s/\s{2,}/ /g;
    $line =~ s/\t/ /g;
    $line =~ s/\+/%/g;
    $line =~ s/\?/%/g;
    $line =~ s/\*/%/g;
    return($line);
}

sub get_full_genelist {
#Creates a list of genes to BLAST from the protein sequence file

    my ($filename, $gene_field) = @_;
    my(@names, $line, $i, $temp);

    open(READFASTA, "<" . $filename) or die "Can't open FASTA file $!\n";
    
    while ($line = <READFASTA>)
    {
	if ($line =~ />.{0,900}/)
	{
	    $line=clean_line($line);
	   
	   
	    push(@names, $line);
	}	
    }
    close (READFASTA);
    return(@names);
}


sub get_file_genelist {
#Creates a list of files to BLAST from a user file

    my ($filename) = @_;
    my(@names, $line);

    open(READPAIRS, "<" . $filename) or die "Can't open partners file $!\n";
    
    while ($line = <READPAIRS>)
    {
	$line =~ s/\sdegree\s\d.{0,200}\n//;
	$line =~ s/\n//;
	push(@names, $line);	
    }
    close (READPAIRS);
    return(@names);
}



sub get_gene_partners {
#Gets the list of interaction partners from the user file

    my($filename, @genes) = @_;
    my(@lines, @degrees, $newdeg, $len, $i, $filelen, $j);
    
    open(READPAIRS, "<" . $filename) or die "Can't open partners file $!\n";
    @lines=<READPAIRS>;
    close (READPAIRS);
    $filelen=@lines;

    $len=@genes;
    for($i=0; $i<$len; $i++) {
	$j=0;
	while (($j < $filelen) && (!($lines[$j] =~ /$genes[$i]\sdegree\s(\d{1,10}).{1,200}\n/))) {$j++;}
	
	if ($j < $filelen) {
	    $lines[$j] =~ /$genes[$i]\sdegree\s(\d{1,10}).{1,200}\n/;
	    $newdeg=$1;
	}
	else {
	    $newdeg="NA";
	}
	push(@degrees, $newdeg);
    }
    return(@degrees);
}


sub find_sequence {
#Finds the sequence for a gene in a FASTA file

    my($genename, $exactgenename, @inputseqs) = @_;
    my(@pro_lines, $line, @retseq, $i, $j, $len, $found, $temp, $header);

    $found = 0;
    $len=@inputseqs;
    $i=0;

   
    while (($i<$len) && ($found == 0))
    { 
	while (($i<$len) && (!($inputseqs[$i] =~ />/))) {$i++;}
	
	if (($inputseqs[$i] =~ />/)) {
	    $line=clean_line($inputseqs[$i]);
	    if (($line =~ /$genename/) || (($exactgenename ne "") && ($line =~ /$exactgenename/))) {
		$found=1;		
		$header=$line;
	    }
	    else {$i++;}
	  
	}
    }

    if ($found == 1) {
	$i++;
	while (($i<$len) && !($inputseqs[$i] =~ />/))
	{
	    $inputseqs[$i] =~ s/\n//; 
	    $inputseqs[$i] =~ s/\n//;
	    $inputseqs[$i] =~ s/\*//;
	    push(@retseq, $inputseqs[$i]);
	    $i++;
	}
    }
    else {
	$retseq[0]="NONE";
    }
    return($header, @retseq);
}

sub find_sequence_num {
#Finds the sequence for a gene in a FASTA file

    my($genename, $exactgenename, @inputseqs) = @_;
    my(@pro_lines, $line, $i, $count,  $j, $len, $found, $temp);

    $found = 0;
    $len=@inputseqs;
    $i=0;
    $count=0;
 
    while (($i<$len) && ($found == 0))
    { 
	while (($i<$len) && (!($inputseqs[$i] =~ />/))) {$i++;}
	
	if (($inputseqs[$i] =~ />/)) {
	    $line=clean_line($inputseqs[$i]);
	    if (($line =~ /$genename/) || (($exactgenename ne "") && ($line =~ /$exactgenename/))) {
		$found=1;		
	    }
	    else {$i++; $count++;}
	  
	}
    }

    print "Analyzed protein sequences: looking for $genename, found: $found at $count\n";
    if ($found == 1) {
	return($count);
    }
    else {
	return(-1);;
    }
    
}

sub find_sequence_w_num {
#Finds the sequence for gene number # in a FASTA file

    my($num, @inputseqs) = @_;
    my(@pro_lines, $count, $line, @retseq, $i, $j, $len, $found, $temp);

    $found = 0;
    $len=@inputseqs;
    $i=0;
    $count=0;

    while (($i<$len) && ($found == 0))
    { 
	while (($i<$len) && (!($inputseqs[$i] =~ />/))) {$i++;}
	

	if (($inputseqs[$i] =~ />/)) {
	    if ($count == $num) {
		$found=1;
	    }
	    else {$i++;}
	    $count++;
	}
    }

    if ($found == 1) {
	$i++;
	while (($i<$len) && !($inputseqs[$i] =~ />/))
	{
	    $inputseqs[$i] =~ s/\n//; 
	    $inputseqs[$i] =~ s/\n//;
	    $inputseqs[$i] =~ s/\*//;
	    push(@retseq, $inputseqs[$i]);
	    $i++;
	}
    }
    else {
	$retseq[0]="NONE";
    }
    return(@retseq);
}

sub write_fasta_file {
    my($outputfile, $genename1, @sequence1, $genename2, @sequence2) = @_;
    my($i, $len);

    open(WRITESEQ, ">" . $outputfile) or die "Can't open output file $outputfile\n";
   
    print WRITESEQ ">$genename1\n";
    $len=@sequence1;
    
    for($i=0; $i<$len; $i++) 
	{
	    print WRITESEQ $sequence1[$i];
	    print WRITESEQ "\n";
	   	    
	}
	
    close(WRITESEQ);
}


sub parse_blastout {
    my($sourcename, $tempseqfile, $take_best_match, $eval_limit, $matrix_loc, $gene_field) = @_;
    my($output, @lines, @partners, $j, $k, $no_matches);
   
    if ($matrix_loc =~ /NONE/) {
	$output= `nice blastp $protein_seq_file $tempseqfile`;
    }
    else {
	$output= `nice blastp $protein_seq_file $tempseqfile -matrix $matrix_loc`;
	
    }
    @lines = split(/\n/,$output);
    $k=0;
    $j=10;
    while (!($lines[$j] =~ /Sequences\sproducing\sHigh-scoring\sSegment\sPairs/))
    {
	$j++;
    }

    $j+=2;

 
    if ($lines[$j+1] =~ /NONE/)
    {
	$no_matches=1;
    }
	
    if (!($no_matches))
    {
	if ($take_best_match =~ /yes/i)
	{
	    @partners=get_best_match($sourcename, $j, $gene_field, @lines);
	}
	else 
	{
	    @partners=get_all_high_matches($sourcename, $eval_limit, $j, $gene_field, @lines);
	}
    }
    else
    {
	push(@partners, "NONE");
    }
    return (@partners);
}



sub get_all_high_matches {
    my($sourcename, $eval_limit, $start, $gene_field, @lines) = @_;
    my($i, $j, $temp, $eval, @partners);

    $i=$start;
#    print "Source: $sourcename\n";
   
    while ($lines[$i] ne "")
    { 
#	print "Original hit: $lines[$i]\n";
	$lines[$i]=clean_line($lines[$i]);

	$lines[$i] =~ /(\s{1,8}\d{1,6}\s{1,2}(\d{1,3}\.\d{0,7}e?-?\d{0,4})\s{1,8}\d{1,6})/;
	
	$temp=$1;
	$eval=$2;
	
	$lines[$i] =~ s/$temp//;
	
	$lines[$i]=~ s/(\.{3})//;
	$lines[$i] =~ s/\s{1,50}$//;
	$eval *= 1.0;
	
#	print "Hit: $lines[$i]: eval: $eval: ";
	
	if (!($sourcename =~ /$lines[$i]/))
	{
#	    print "Hit: $lines[$i]: eval: $eval: ";
	    if ($eval <= $eval_limit )
	    {   
#		print "...Accepted\n";
		push(@partners, $lines[$i]);
	    }
#	    else {
#		print "...Rejected\n";
#	    }
	}
#else	{print "IDENTICAL\n";}
	$i++;
    }
    return(@partners);
}



sub get_best_match {
    my($sourcename, $start, $gene_field, @lines) = @_;
    my($i, $j, $temp, @partners);

    $i=$start;
  
    while ($sourcename =~ /$lines[$i]/)
    {
	$i++;
    }

    $lines[$i]=clean_line($lines[$i]);
    $lines[$i] =~ /(\s{1,8}\d{1,6}\s{1,2}(\d{1,3}\.\d{1,7}e?-?\d{0,4})\s{1,8}\d{1,6})/;
    
    $temp=$1;
   
    $lines[$i] =~ s/$temp//;
    
    $lines[$i]=~ s/(\.{3})//;
    $lines[$i] =~ s/\s{1,30}$//;
    
    
   
    push(@partners, $lines[$i]);
    
    return(@partners);
}


sub get_seqlen {
    my(@seq) = @_;
    my($i, $len, $count, $mashline, @line);

    $count=0;
    $len=@seq;
    
    for($i=0; $i<$len; $i++) 
    {
	$mashline=$seq[$i];
	$mashline =~ s/\n//;
	@line=split //, $mashline;
	$count += @line;

    }
    return($count);
}



sub parse_clustal_output {
    my($filename) = @_;
    my($i, $output, @lines, $len, $newfile);

    $newfile=$tempseqfile;
    $newfile =~ s/fas/pir/;

    if (-e $newfile) {
	`rm $newfile`; }

    $output=`nice clustalw $tempseqfile /output=pir`;
 
    @lines = split(/\n/,$output);
    $len=@lines;
    $i=0;

   
    if (-e $newfile) {
	return(1);
    }
    else {
	return(0);
    }
}



sub find_percent_diff {
    my($filename, $pro_seq) = @_;
    my($output, @lines, $i, $len, $percent_sim, $newfile, $num_algn_res);

  
    $output=`calc_percent_diff $filename -d:$pro_seq -nogap`;
  
    @lines = split(/\n/,$output);
   
    $i=0;
    if (!($lines[$i] =~ /Number of non-gap aligned residues/)){
	return(-1);
    }
    else {
	while (!($lines[$i] =~ /gene1/i)) {
	    if ($lines[$i] =~ /Number\sof\snon-gap\saligned\sresidues:\s(\d{1,30})/)
	    {
		$num_algn_res=$1;
	    }
	    $i++;
	}
	
	$lines[$i] =~ /gene1\s{1,12}--\s{1,12}(\d\.\d{1,10})/i;
	$percent_sim= $1;
	
	return(1.0-$percent_sim, $num_algn_res);
}
    
}




sub nuc_align {
  			    
    my($nuc_align, $pro_align, $paramfile, $genecode_arg) = @_;
    my($output, @lines, @lines2, $i, $len, $len2,$error);

    $output = `nice algndna_new $pro_align $nuc_align XXX -f:$paramfile $genecode_arg`;
   
	
    @lines = split(/\n/, $output);
    $len=@lines;

   
    $error=0;
    if (-e  $nuc_align) {
	open(TESTALGN, "<" . $nuc_align);
	@lines2=<TESTALGN>;
	$len2=@lines2;
	if ($len2>2)
	{
	    for ($i=0; $i<$len; $i++)
	    {
		if ($lines[$i] =~ /ERROR:\sAligned\sDNA\ssequence\sdoesn\'t\smatch\sorignal\sprotein\ssequence/)
		    {
			$error=-1;
		    }
	    }
	}
	else {
	    $error=-2;
	}
    }
    else {
	print "Error: alignment file doens't exist\n";
	$error=-3;
    }
    return($error);
}


sub nuc_align_error {
    my ($error)=@_;
    my ($string_error);

    if($error == -1) {$string_error= "ERROR: Back-translation does not match input protein alignment\n";}
    if ($error == -2) {$string_error= "ERROR: Nucleotide alignment file truncated\n";}
    if ($error == -3) {$string_error= "ERROR: Nucleotide alignment file does not exist\n";}
    return($string_error);
}

sub find_ks_ka {
    my($like_out)=@_;
    my($i, $j, @lines, $len, $ks, $ka, $sat);

    $ks="NA";
    $sat="";


    @lines = split(/\n/,$like_out);
    $len=@lines;
    $i=0;
    while (!($lines[$i] =~ /Name\sdS\sdN/) && $i<$len) {
	$i++;
    }

    for($j=0; $j<3; $j++)
    {
       
	$lines[$i+$j] =~ /\w{1,30}\s(\d{1,5}\.\d{1,12}e?-?\d{0,4})\s(\d{1,5}\.\d{1,12}e?-?\d{0,4})/;
	    if ($1 != 0) {
		$ks=$1;
		$ka=$2;
	    }
	    
    }

    while($i<$len) {
	if ($lines[$i] =~ /WARNING:  Synonymous substitutions between these two sequences are likely saturated/) {
	    $sat="*";
	}

	$i++;
    }
    return($ks, $ka, $sat);
}



