//Copyright 1999-2002 Gavin Conant

#include <fstream.h>
#include <iostream.h>
#include "maxlike.h"
#include "codon_like.h"
#include "exchange.h"
#include "tree.h"
#include "read_tree.h"
#include "read_seq.h"
#include "gen_code.h"
#include "powell.h"
#include "pair_dist.h"
#include "gen_dna_funcs.h"
#include <string.h>



void parse_args(int argc, char **argv, BOOL &removegaps, BOOL &codon_freqs,  BOOL &fixed_freqs, BOOL &fixed_pns, Exchange *curr_exchange, double &tol);


int main(int argc, char *argv[])
{
  int i=0, j,k, ntaxa, nchars, temp_codon[3];
  double p_ns, use_ks, use_ka, codon_basefreqs[3][4], tol, old_b1, old_b2, old_ka, old_ks, old_kaks, sat_lnL, unsat_lnL;
  BOOL codon_freqs, removegaps, fixed_freqs, interleave, fixed_pns;

  //Classes
  Exchange current_exchange;
  Tree *current_tree;
  Two_Taxa_Tree get_tree;
  Read_Sequence *read_seqs;
  Sequence_dataset *nuc_data, *new_data, *current_data, *third_pos_data;
  Genetic_code *current_genetic_code;
  Like_model *model;  
  Powell current_powell;
  JC_dist *backup_dist;
  L_93_Synon_dist *syn_dist;
  L_93_Non_Synon_dist *nonsyn_dist;
  if (argc>1)
    {  
      parse_args(argc, argv, removegaps, codon_freqs, fixed_freqs, fixed_pns, &current_exchange, tol);

      current_genetic_code=create_genetic_code(&current_exchange);

      if (current_exchange.curr_groups == 0) {
	if (fixed_freqs==TRUE) 
	  current_exchange.set_model(MG_94_K2P);                            //Muse and Gaut 94 model
	else
	  current_exchange.set_model(MG_94_HKY);
	}
      else {
	if (fixed_freqs==TRUE) 
	  current_exchange.set_model(C_00_K2P);                            //Groups model
	else
	  current_exchange.set_model(C_00_HKY);
      }

      
   
    //Guesses dataformat
      switch(guess_dataformat(current_exchange.get_datafile(), strlen(current_exchange.get_datafile())))
	{
	case NEXUS:
	  read_seqs=new Read_Nexus;
	  break;
	case PIR:
	  read_seqs=new Read_PIR;
	  break;
	case PHYLIP:
	  interleave=test_interleave(current_exchange.get_datafile());

	  if (interleave == TRUE)
	    read_seqs=new Read_Phylip_interleave;
	  else
	    read_seqs=new Read_Phylip_noninterleave;
	  break;
	case FASTA:
	  read_seqs=new Read_FASTA;
	  break;
	} 

    //Reads the data in the selected format
    nuc_data= read_seqs->get_dataset(ntaxa, nchars, current_exchange.get_datafile(), FALSE);
    if (nuc_data==0)
      {
	delete read_seqs;
	return(-1);
      }

    
    current_exchange.set_num_taxa(ntaxa);
    current_exchange.set_num_sites(nchars);
    current_exchange.set_trs_trv(1.0);
    current_exchange.set_num_rates(1);
    current_exchange.set_num_p_nonsyn(1);
 
    if(removegaps == TRUE) {
      new_data= remove_gaps (nuc_data, FALSE);
      delete nuc_data;
      nuc_data=new_data;
      nchars=(*nuc_data)[0].Sequence_size();
      current_exchange.set_num_sites(nchars);
    }
    
    if (fixed_freqs==FALSE)
      {
	//If we use non-equal basefreqs, we tell the optimizer not to optimize them
	current_exchange.fix_basefreq();

      
	if (codon_freqs==FALSE) {
	  for(i=0; i<4; i++) 
	    current_exchange.set_basefreqs(observed_basefreqs (nuc_data, i), i);  
	  //cout<<"Using base freqs: A: ";
	  //cout<<current_exchange.return_basefreq(0)<<" C: "<<current_exchange.return_basefreq(1)<<
	  //  " G: "<<current_exchange.return_basefreq(2)<<" T: "<<current_exchange.return_basefreq(3)<<endl;
	}
	else {
	  for(i=0; i<3; i++)
	    for(j=0; j<4; j++)
	      codon_basefreqs[i][j]= observed_codon_basefreqs(nuc_data, i, j);

	  current_exchange.set_use_codon_basefreqs(codon_basefreqs);
	  //cout<<"Using codon freqs: ";
	  // cout<<"Position 1: A: "<<current_exchange.return_codon_basefreq(0,0)<<" C: "<<current_exchange.return_codon_basefreq(0,1)<<
	  //  " G:"<<current_exchange.return_codon_basefreq(0,2)<<" T: "<<current_exchange.return_codon_basefreq(0,3)<<endl;
	  //cout<<"Position 2: A: "<<current_exchange.return_codon_basefreq(1,0)<<" C: "<<current_exchange.return_codon_basefreq(1,1)<<
	  //  " G: "<<current_exchange.return_codon_basefreq(1,2)<<" T: "<<current_exchange.return_codon_basefreq(1,3)<<endl;
	  //cout<<"Positon 3: A: "<<current_exchange.return_codon_basefreq(2,0)<<" C: "<<current_exchange.return_codon_basefreq(2,1)<<
	  //  " G: "<<current_exchange.return_codon_basefreq(2,2)<<" T: "<<current_exchange.return_codon_basefreq(2,3)<<endl;

	}
      }
    
    
   
    syn_dist = new L_93_Synon_dist(nuc_data, &current_exchange);
    nonsyn_dist = new L_93_Non_Synon_dist(nuc_data, &current_exchange);
    syn_dist->find_dist();
    nonsyn_dist->find_dist();
    
    //cout<<"Ks: "<<syn_dist->return_dist(0,1)<<"\nKa: "<<nonsyn_dist->return_dist(0,1)<<endl;
    
    if (nonsyn_dist->return_dist(0,1)>0 && nonsyn_dist->return_dist(0,1)<10.0)
      use_ka=nonsyn_dist->return_dist(0,1);
    else
      use_ka=0.1;
    
    if (syn_dist->return_dist(0,1)>0 && syn_dist->return_dist(0,1)<=10.0)
      use_ks=syn_dist->return_dist(0,1);
    else                                  //Li's method failed--use the Jukes-Cantor distance on 3rd positions to estimate Ks
      {
	third_pos_data=new Sequence_dataset(2, (*nuc_data)[0].Sequence_size()/3);
	j=0;
	for (i=0; i<(*nuc_data)[0].Sequence_size(); i++){
	  if (i%3 == 2){
	    (*third_pos_data)[0].Assign_site(j,  (*nuc_data)[0][i]);
	    (*third_pos_data)[1].Assign_site(j,  (*nuc_data)[1][i]);
	    j++;
	  }
	}
	backup_dist = new JC_dist(third_pos_data, &current_exchange);
	backup_dist->find_dist();
	use_ks=backup_dist->return_dist(0,1);
	if (use_ks < 1e-7)
	  use_ks=1e-7;
	else if (!((use_ks > 0.0) && (use_ks < 10.0)))
	  use_ks=1.0;
      }
    
    p_ns=use_ka/use_ks;
    

    //cout<<"Setting initial Ks: "<<use_ks<<" Initial Ka: "<<use_ka<<endl;
    //cout<<"Setting p_non_syn: "<<p_ns<<endl;
    if (fixed_pns==FALSE) 
      current_exchange.set_p_non_syn(0, p_ns);
    else {
      cout<<"Fixing Ka/Ks at 1.0\n";
      current_exchange.fix_p_non_syn(0, 1.0);
    }

    current_exchange.set_p_inter_group(0, p_ns);
    
    //Gets the tree
    current_tree=get_tree.create_two_taxa_tree(&current_exchange, use_ks);

    //Translated the nucleotide data into codes for 61 non-stop codons
    current_data= current_genetic_code->make_codon_sequence (nuc_data); 
    
 
    //Sets up the likelihood model class  
    switch (current_exchange.get_model())
      {
      case MG_94_K2P:
	model= new MG_94_K2P_model (&current_exchange, current_data, current_tree,
				    current_genetic_code);
	break;
      case MG_94_HKY:
	model= new MG_94_HKY_model (&current_exchange, current_data, current_tree,
				    current_genetic_code);
	break;
      case C_00_K2P:
	model= new C_00_K2P_model (&current_exchange, current_data, current_tree,
				   current_genetic_code);
	break; 
      case C_00_HKY:
	model= new C_00_HKY_model (&current_exchange, current_data, current_tree,
				   current_genetic_code);
	break;
      }
    
    if (tol != 0)
      current_powell.set_tolerance(tol);
    
    cout<<"Begining global optimization\n"<<flush;
    current_exchange.set_saved_lnL(current_powell.Init_min(model, &current_exchange, TRUE));
    
    model->newton_branch_opt();
    unsat_lnL= current_exchange.get_saved_lnL();

    model->set_expect_subs();
    cout<<"Transition/Transversion ratio: "<<model->get_obs_trs_trv_ratio()<<endl;
    cout<<"Name\tdS\tdN\n";
    for(i=0; i<current_exchange.get_num_branches(); i++)
      {
	cout<<(*current_tree)[i]->get_name()<<"\t"
	    <<(*current_tree)[i]->expect_subs_site()<<"\t"
	    <<fabs((*current_tree)[i]->expect_nsyn_subs_site())<<endl; 
      }

  
    old_b1=(*current_tree)[0]->get_brnlen();
    old_b2=(*current_tree)[1]->get_brnlen();
    old_kaks=current_exchange.get_p_non_syn(0);


    if (old_b1 != 0) {
      old_ka=(*current_tree)[0]->expect_nsyn_subs_site();
      old_ks=(*current_tree)[0]->expect_subs_site();
      current_exchange.set_p_non_syn(0, old_ka/(10*old_ks));				     
      
      model->set_ratios((*current_tree)[0]);

      (*current_tree)[0]->set_brnlen(10*old_ks/(*current_tree)[0]->get_syn_ratio());
    }   
    else {
      old_ka=(*current_tree)[1]->expect_nsyn_subs_site();
      old_ks=(*current_tree)[1]->expect_subs_site();
      current_exchange.set_p_non_syn(0, old_ka/(10*old_ks));				     
      
      model->set_ratios((*current_tree)[1]);

      (*current_tree)[1]->set_brnlen(10*old_ks/(*current_tree)[1]->get_syn_ratio());     
    }
    
    if (old_ks > MIN_BRLEN) {
      model->recalculate_transprobs();
      sat_lnL=model->find_ln_like_w_rates();
      
      cout<<"Sat lnL: "<<sat_lnL<<endl;
      cout<<"Unsat lnL: "<<unsat_lnL<<endl;
      
      if (sat_lnL>=unsat_lnL)
	cout<<"WARNING:  Synonymous substitutions between these two sequences are likely saturated.  Sat lnL: "<<sat_lnL<<" Unsat lnL: "<<unsat_lnL<<"\n";
    
    }  
      (*current_tree)[0]->set_brnlen(old_b1);
      (*current_tree)[1]->set_brnlen(old_b2);
      current_exchange.set_p_non_syn(0, old_kaks);
      
      model->recalculate_transprobs();
      sat_lnL=model->find_ln_like_w_rates();
    




    delete syn_dist;
    delete nonsyn_dist;
    delete model;
    delete current_tree;
    delete nuc_data;
    delete current_data;
    delete read_seqs;
    delete current_genetic_code;

    return(0);
  } 
  else
    {
      cerr<<"Usage: like_pair_dist <sequence file> (-fixKa/Ks) (-nogap) (-obs_freqs)/(-codon_freqs) (-v:(U/VM/YM/MM/IM/CM/EM) (-g # groups <groups_file>\n";
      cerr<<"-v sets the genetic code: \n"<<"----------------------\n";
	cerr<<"U =Universal (default)\nVM=Vert. Mitochondrial\nYM=Yeast Mito.\nMM=Mold Mito.\nIM=Invert. Mito.\nCM=Ciliate Nuclear\nEM=Echino. Mito\n";
     
      return(-1);
    }
}  //End main





void parse_args(int argc, char **argv, BOOL &removegaps, BOOL &codon_freqs, BOOL &fixed_freqs, BOOL &fixed_pns, Exchange *curr_exchange, double &tol)
{
  int i, j, groups;
  char datafile[100], groups_file[100], tol_string[30];
  removegaps=FALSE;
  codon_freqs=FALSE;
  fixed_freqs=TRUE;
  fixed_pns=FALSE;
  curr_exchange->set_genetic_code(UNIVERSAL);
  

  tol=0;
  strcpy(datafile, argv[1]);
  curr_exchange->set_datafile(datafile);
  i=2;
  while (i<argc)
    {
      switch(argv[i][1]) {
      case 'n':
      case 'N':
	removegaps=TRUE;
	i++;
	break;
      case 'c':
      case 'C':
	fixed_freqs=FALSE;
	codon_freqs=TRUE;
	i++;
	break;
      case 'o':
      case 'O':
	fixed_freqs=FALSE;
	i++;
	break;
      case 'v':
      case 'V':
	switch (argv[i][3]) {
	case 'u':
	case 'U':
	  curr_exchange->set_genetic_code(UNIVERSAL);
	  break;
	case 'v':
	case 'V':
	  curr_exchange->set_genetic_code(VERT_MITO);
	  break;
	case 'y':
	case 'Y':
	  curr_exchange->set_genetic_code(YEAST_MITO);
	  break;
	case 'm':
	case 'M':
	  curr_exchange->set_genetic_code(MOLD_MITO);
	  break;
	case 'i':
	case 'I':
	  curr_exchange->set_genetic_code(INVERT_MITO);
	  break;
	case 'c':
	case 'C':
	  curr_exchange->set_genetic_code(CILIATE_NUC);
	  break;
	case 'e':
	case 'E':
	  curr_exchange->set_genetic_code(ECHINO_MITO);
	  break;
	}
	i++;
	break;
      case 'g':
      case 'G':
	curr_exchange->set_model(C_00_K2P);
	groups=string_to_int(argv[i]);
	strcpy(groups_file, argv[i+1]);
	
	curr_exchange->curr_groups=new Amino_acid_group(groups, groups_file);
	i+=2;
	break;
      case 't':
      case 'T':
	j=3;
	cout<<"ARg: "<<argv[i]<<endl;
	while(j<strlen(argv[i])) 
	  tol_string[j-3]=argv[i][j++];
	tol=string_to_float(tol_string);
	i++;
	break;

      case 'f':
      case 'F':
	fixed_pns=TRUE;
	i++;
	break;
	
      }
    }

}









