//Copyright 1999-2002 Gavin Conant

#include "tree.h"
#include "iostream.h"
#include <math.h>
#include <string.h>
#include <float.h>

//#define DEBUG


Branch::Branch()
{
  cerr<<"Call to default constructor of Branch class\n";
}

Branch::Branch(Exchange *cexchange)
{
	int i,j;
  
	curr_exchange=cexchange;

	strcpy(name, "NONE");
	tip=FALSE;
	taxa_id=-1;
	uninitialized=TRUE;
	basefreqs[0]=basefreqs[1]=basefreqs[2]=basefreqs[3]=0.25;
	brlen=0.0;
	expect_numsubs_site=0.0;
	parent=children[0]=children[1]=sibling=0;
	pns_num=0; 
	pitg_num=0;
	nonsyn_pattern=0;
	which_nonsyn_patt=&Branch::pns_num;

	for(i=0; i<curr_exchange->get_num_aa_props()+1; i++)
		aa_prop_num[i]=0;

	
	if(curr_exchange->get_num_matrices() != 0)
		matrix_coeff_nums=new int[curr_exchange->get_num_matrices()];
	else
		matrix_coeff_nums=0;


	transprobs=new double**[curr_exchange->get_num_rates()];
	transprobs_prime=new double**[curr_exchange->get_num_rates()];
	transprobs_double_prime=new double**[curr_exchange->get_num_rates()];

	condprobs=new double [curr_exchange->get_condlike_size()];
	for(i=0; i<curr_exchange->get_num_rates(); i++)
	{
		transprobs[i]=new double*[curr_exchange->get_condlike_size()+1];
		transprobs_prime[i]=new double*[curr_exchange->get_condlike_size()];
		transprobs_double_prime[i]=new double*[curr_exchange->get_condlike_size()];

		  for(j=0; j<curr_exchange->get_condlike_size()+1; j++)
			transprobs[i][j]=new double [curr_exchange->get_condlike_size()+1];
		  for (j=0; j<curr_exchange->get_condlike_size(); j++) {
			  transprobs_prime[i][j]=new double[curr_exchange->get_condlike_size()];
			  transprobs_double_prime[i][j]=new double[curr_exchange->get_condlike_size()];
			}
    }


	//Hold derivatives of the transition probablities
	
	

}  //End Branch::Branch(Exchange *cexchange)



Branch  & Branch::operator=(Branch & assign_from)
{
  int i, j, k;
 
  set_name(assign_from.get_name());
 
  brlen=assign_from.get_brnlen();
  expect_numsubs_site=assign_from.expect_subs_site();
  
  tip=assign_from.is_tip();
  if (tip == TRUE) 
	  children[0]=children[1]=0;
  
  taxa_id=assign_from.get_taxa_id();
  pns_num=assign_from.get_p_nonsyn_num();
  pitg_num=assign_from.get_p_intergroup_num();
  for(i=0; i<curr_exchange->get_num_aa_props()+1; i++)
	  aa_prop_num[i]=assign_from.get_aa_prop_num(i);

  for(i=0; i<curr_exchange->get_num_rates(); i++)
      for(j=0; j<curr_exchange->get_condlike_size()+1; j++)
          for(k=0; k<curr_exchange->get_condlike_size()+1; k++)
             transprobs[i][j][k]=assign_from.get_trpb(i,j,k);
  //Don' think this is right--MS VC complains w/o it
  return(*this);
}



void Branch::set_aa_prop_num(int aanum, AA_PROPERTIES prop)
{
	aa_prop_num[curr_exchange->get_prop_index_num(prop)]=aanum;
}

 
void Branch::set_matrix_coeff_num(int matrix, int matrix_num)
{
	if (matrix_coeff_nums != 0)
		matrix_coeff_nums[matrix]=matrix_num;
	else
		cerr<<"Error: request to set matrix coeff when matrices are not in use\n";

}

void Branch::set_nonsyn_pattern(int pat)
{
	which_nonsyn_patt=&Branch::nonsyn_pattern;
	nonsyn_pattern=pat;

}


double Branch::get_trpb(int rate, int start, int end)
{
  if (end<curr_exchange->get_condlike_size())
    return(transprobs[rate][start][end]);
  else
    return(transprobs[rate][start][curr_exchange->get_condlike_size()]);
} //End Branch::get_trpb


int Branch::get_aa_prop_num(AA_PROPERTIES prop)
{
	return(aa_prop_num[curr_exchange->get_prop_index_num(prop)]);
}


int Branch::get_aa_prop_num(int prop)
{
	return(aa_prop_num[prop]);
}



int Branch::get_matrix_coeff_num(int matrix)
{
	if (matrix_coeff_nums != 0)
		return(matrix_coeff_nums[matrix]);
	else
		cerr<<"Error: request to get matrix coeff when matrices are not in use\n";
}



Branch * Branch::get_child(int childnum)
{
  if (childnum==0)
    return(children[0]);
  else
    return(children[1]);
}




void Branch::set_child(Branch *newchild, int childnum)
{
  if (childnum==0)
    children[0]=newchild;
  else
    children[1]=newchild;
}


void Branch::null_child(int childnum)
{
  if (childnum==0)
    children[0]=0;
  else
    children[1]=0;
}



Branch::~Branch()
{
  int i, j;

  
  for(i=0; i<curr_exchange->get_num_rates(); i++)
    {
      for(j=0; j<curr_exchange->get_condlike_size()+1; j++)
		delete[] transprobs[i][j];
	  for(j=0; j<curr_exchange->get_condlike_size(); j++) {
			delete[] transprobs_prime[i][j];
			delete[] transprobs_double_prime[i][j];
		}

      delete[] transprobs[i];
	  delete[] transprobs_prime[i];
	  delete[] transprobs_double_prime[i];
    }
  
 
  


  delete[] transprobs;
  delete[] condprobs;
  delete[] transprobs_prime;
  delete[] transprobs_double_prime;

  if(matrix_coeff_nums != 0)
	  delete[] matrix_coeff_nums;

}





Tree::Tree()
{
  cerr<<"Call to default constructor of Tree class\n";
}



Tree::Tree(Exchange *cexchange)
{
  int i;
  curr_exchange=cexchange;

  //We flag that this tree contains its own local data
  tree_is_local_mem=TRUE;
   null_root_brlns=TRUE;	

	start_extra_brn=new Branch_list;
	extra_brn=start_extra_brn;
	start_tips=new Branch_list;
	tips=start_tips;

	start_tips->next=0;
	start_extra_brn->next=0;

  if (curr_exchange->get_num_taxa() <= 3)
    three_taxa_tree=TRUE;
  else
    three_taxa_tree=FALSE;

  prune_root=0;
  
  //Declares a new array of branches large enough for the tree
  tree = new Branch * [curr_exchange->get_num_branches()];
  
  for (i=0; i<curr_exchange->get_num_branches(); i++)
    tree[i]=new Branch(curr_exchange);
}



Tree::Tree(Exchange *cexchange, Branch *new_root)
{
   int i;
   curr_exchange=cexchange;

   //This tree uses memory allocated in another tree object--
   //We won't dealloc it in the destructor
   tree_is_local_mem=FALSE;
	null_root_brlns= FALSE;
	start_tips=0;
	start_extra_brn=0;

   prune_root=0;
   

   root=new_root;
   if (new_root->is_tip() == FALSE) {
     if (!((root->get_child(0)->is_tip() ==TRUE) && (root->get_child(1)->is_tip() ==TRUE))) {
       set_root(root);
       if ((find_null_branch()->get_child(0)->is_tip() ==TRUE) &&
	   (find_null_branch()->get_child(0)->is_tip() ==TRUE) && find_null_branch()->get_sibling()->is_tip()==TRUE)
	 three_taxa_tree=TRUE;
       else
	 three_taxa_tree=FALSE;
     }
   }
  
}



Branch * Tree::operator[] (int element)
{
  if (element<curr_exchange->get_num_branches())
    return(tree[element]);
  else
    {      cerr<<"Invalid Branch number\n";
      return(dummy);
    }  
}



Tree & Tree::operator= (Tree & assign_from)
{
  int i, j;

  for (i=0; i<curr_exchange->get_num_branches(); i++)
  {
	
    *(tree[i])=*(assign_from[i]);
    
    if (assign_from[i]!=assign_from.find_root())
    {     
      j=0;
      
      while (assign_from[j]!=assign_from[i]->get_sibling()) j++;
      set_as_siblings(tree[i], tree[j]);
      if (tree[i]->is_tip()==FALSE)
      {	
	j=0;
	
	while (assign_from[j]!=assign_from[i]->get_child(0)) j++;
	set_as_parent_child(tree[i], tree[j], 0);
	j=0;

	while (assign_from[j]!=assign_from[i]->get_child(1)) j++;
	set_as_parent_child(tree[i], tree[j], 1);
      }
      j=0;
      
      while (assign_from[j]!=assign_from[i]->get_parent()) j++;
      if (assign_from[i]->get_parent()->get_child(0)==assign_from[i])
          set_as_parent_child(tree[j], tree[i], 0);
      else
          set_as_parent_child(tree[j], tree[i], 1);
    }
   
    else
    {
      j=0;
	
      while (assign_from[j]!=assign_from[i]->get_child(0)) j++;
      set_as_parent_child(tree[i], tree[j], 0);
      j=0;
      
      while (assign_from[j]!=assign_from[i]->get_child(1)) j++;
      set_as_parent_child(tree[i], tree[j], 1);
      tree[i]->null_parent();
      tree[i]->null_sibling();
    }
  }
  set_root(tree[0]);
  //Don't think this is right--MS VC complains w/o it
  return(assign_from);
}



void Tree::describe_tree()
{
  //Prints out the siblings of all of the tip nodes in the tree array structure

  for (int i=0; i<curr_exchange->get_num_branches(); i++)
    if (tree[i]!=root)
      cout<<tree[i]->get_name()<<" "<<" (Sibling: "<<tree[i]->get_sibling()->get_name()<<"): " 
	 <<" Len: "<<tree[i]->get_brnlen()<<endl;
    else
      cout<<"ROOT: "<<tree[i]->get_name()<<" Len: "<<tree[i]->get_brnlen()<<endl;
  
}  //End Tree::describe_tree




Branch * Tree::get_pott_branch_num(Branch *start, int &curr_num, int target_num)
//This function finds the branch with post-order tree transversal number ==
//target_num + (original curr_num +1).  Thus, if called from the tree root with
//curr_num == -1, it will find the target_num +1 th branch in the post-order transversal
{
  Branch *ret=0;

 
  //First progress to the left tip
  if (start->is_tip() == FALSE)
    ret=get_pott_branch_num(start->get_child(0), curr_num, target_num);
 
  //Have we already found the branch--then return
  if (ret != 0)
    return(ret);
  
  //Number this branch
  curr_num++;
 
  //This branch is the target
  if (curr_num == target_num)
    return(start);

  //Check if this branch is a left child (root check performed to avoid
  //seg fault if check parent of root)
  if ((start != root) && (start->get_parent()->get_child(0)==start))
    {
      //If sib is tip, number it and check if it is the target
      if (start->get_sibling()->is_tip() == TRUE)
	{
	  curr_num++;
	   if (curr_num == target_num)
	     return(start->get_sibling());
	   else
	     return(0);
	}
      //If our sib is not a tip, number that subtree
      else
	return(get_pott_branch_num(start->get_sibling(), curr_num, target_num));
    }
  //If we are at a right child just return
  else
    return(ret);
}



int Tree::count_branches_above(Branch *start)
{
  int retval=0;
  if (start->is_tip() == FALSE)
    return(1+count_branches_above(start->get_child(0))+
      count_branches_above(start->get_child(1)));
  else
    return(1);
}


Branch * Tree::tree_bisect_reconnect(Branch * split_loc, int tree_1_join, int tree_2_join, int base_tree)
  //Does a tree bisection and reconnection operation.  Split occurs at split_loc.  Post-oder tree numbering is used to identify the two branches numbered tree_1_join and tree_2_join.  Tree_1_join is the "smaller" of the trees 
  //(i.e. the tree "above" split loc)
{
  int join1num=-1, join2num=-1;
  Branch *join1, *join2, *retval=0;
  Tree *newtree;

  null_root_brlns=FALSE;
  newtree=split_tree(split_loc);

  if (newtree != 0) {
   
    join1 = newtree->get_pott_branch_num(newtree->find_root(), join1num, tree_1_join);
    join2 = get_pott_branch_num(root, join2num, tree_2_join);
        
    if ((join1 == 0) || (join2 == 0)) 
      cerr<<"Tree bisection and reconnection failed: invalid join point on one or both subtrees\n";
    else
      retval=join_trees(newtree, this, join1, join2, base_tree);
    
    delete newtree;
  }
  null_root_brlns=TRUE;
  if(retval != 0)
    set_root(retval);
  return(retval);
}




void Tree::re_root_tree(Branch * root_from)
{
  int root_from_child_num;


  if (root_from != root) {
    if (root->is_tip() != TRUE)
      if (!((root->get_child(0)->is_tip() ==TRUE) && (root->get_child(1)->is_tip() ==TRUE))) {  
	if (three_taxa_tree == TRUE)
	  {
	    if (root_from != find_null_branch()->get_sibling())
	      {
		old_sibling=root_from->get_sibling();
		old_parent=root_from->get_parent();
		single_branch=root->get_child(1);
		
		if(old_parent->get_child(0)==root_from)
		  root_from_child_num=0;
		else
		  root_from_child_num=1;
		
		set_as_parent_child(root, root_from, 1);
		set_as_siblings(root_from, find_null_branch());
		
		set_as_parent_child(old_parent, single_branch, root_from_child_num);
		set_as_siblings(single_branch, old_sibling);
	      }
	  }
	else {
	  if (find_null_branch_id()->get_sibling()!=root_from)
	    {
	      single_branch=root->get_child(1);
	      pair_branch1=find_null_branch()->get_child(0);
	      pair_branch2=find_null_branch()->get_child(1);
	      old_sibling=root_from->get_sibling();
	      old_parent=root_from->get_parent();
	      
	      
	      find_null_branch()->get_child(0)->null_parent();
	      find_null_branch()->get_child(1)->null_parent();
	      find_null_branch()->get_sibling()->null_parent();
	      
	      set_as_parent_child(root, root_from, 1);
	      set_as_siblings(root_from, find_null_branch());
	      
	    
	      if (old_parent == find_null_branch()  )
		{
		
		  set_as_parent_child(find_null_branch(), old_sibling, 0);
		  set_as_parent_child (find_null_branch(), single_branch, 1);
		  set_as_siblings(old_sibling, single_branch);
		}
	      
	      
	      else
		{
		  
		  current=invert_clade(find_null_branch(), old_parent, old_sibling);
		  
		  if (current==single_branch)
		    {
		      set_as_parent_child(current, pair_branch1, 0);
		      set_as_parent_child(current, pair_branch2, 1);
		    }
		  
		  else
		  {
		    if (current==pair_branch1)
		      set_as_parent_child(current, pair_branch2, 0);
		    else
		      set_as_parent_child(current, pair_branch1, 0);
		    set_as_parent_child(current, single_branch, 1);
		  }
		  
		  set_as_siblings(current->get_child(0), current->get_child(1));
		}
	    }
	  set_root(root_from);
	  
	
	}
	diganose_tree(root);
      }
  }
  

}  


void Tree::set_root(Branch *start)
  //This function determines the root of the tree, given some branch on the tree (start)
{
  Branch *loc, *new_null;
 
  loc=start;

 

  while(loc->get_parent()!=0)
    loc=loc->get_parent();


  root=loc; 
 
  if (root->is_tip() == FALSE) {

    if (!((root->get_child(0)->is_tip() == TRUE) && (root->get_child(1)->is_tip() == TRUE)))
      {
	if (find_null_branch_id()->is_tip()==TRUE)
	  {
	    new_null=find_null_branch_id()->get_sibling();
	    set_as_parent_child(root, find_null_branch_id(), 1);
	    set_as_parent_child(root, new_null, 0);
	  }
	
	if (null_root_brlns == TRUE) {
	 
	  find_null_branch_id()->get_sibling()->set_brnlen(find_null_branch_id()->get_sibling()->get_brnlen()+
							   find_null_branch_id()->get_brnlen());
	  
	  find_null_branch_id()->get_sibling()->set_expect_subs_site(find_null_branch_id()->get_sibling()->expect_subs_site()+
								     find_null_branch_id()->expect_subs_site());
	  
	  find_null_branch_id()->set_brnlen(0.0);
	  find_null_branch_id()->set_expect_subs_site(0.0);      

	 
	}	
      }
    //Moves root branch length to one of it's children
    if (((root->get_brnlen() != 0.0) || (root->expect_subs_site() != 0.0)) && (null_root_brlns == TRUE)) {
      find_null_branch_id()->get_sibling()->set_brnlen(find_null_branch_id()->get_sibling()->get_brnlen()+
						       root->get_brnlen());
      find_null_branch_id()->get_sibling()->set_expect_subs_site(find_null_branch_id()->get_sibling()->expect_subs_site()+
								 root->expect_subs_site());
      
      root->set_brnlen(0.0);
      root->set_expect_subs_site(0.0);
      
    }
    while (loc->is_tip()==FALSE)
      loc=loc->get_child(0);
    
    left_tip=loc;
  
  }

}




void Tree::set_rooted_tree(BOOL rooted)
{
  is_rooted_tree=rooted;
}



Branch * Tree::get_leftmost_tip()
{
  return(left_tip);
}




Branch * Tree::find_left_tip(Branch * start)
{
  while (start->is_tip()==FALSE)
    start=start->get_child(0);
  return(start);
}


void Tree::prune_and_reconnect(Branch *subtree_root, Branch *reconnect_loc, double split_fac)
{
	int i, child_num;	
	Branch *old_par, *old_sib, *left, *right, *reconnect_sib, *reconnect_par, *orig_null=0;
	extra_brn=start_extra_brn;
	tips=start_tips;
	diganose_tree(root);
	get_tips(subtree_root, subtree_root);
	tips=start_tips;
	while (tips->element != 0) {
		i=0;
		while (tree[i]->get_taxa_id() != tips->element->get_taxa_id()) i++;

		if (tree[i]->get_parent()->get_parent() != 0) {
			if ((reconnect_loc==tree[i]) || (reconnect_loc==tree[i]->get_parent()))
				reconnect_loc=tree[i]->get_parent()->get_parent();
			add_to_branch_list(extra_brn, tree[i]);
			add_to_branch_list(extra_brn, tree[i]->get_parent());
			old_par=tree[i]->get_parent()->get_parent();
			old_sib=tree[i]->get_sibling();

			if (old_par->get_child(0)==tree[i]->get_parent())
				child_num=0;
			else
				child_num=1;
			set_as_parent_child(old_par, old_sib,child_num);
			set_as_siblings(old_par->get_child(0), old_par->get_child(1));
		}
		else {
			if ((reconnect_loc==tree[i]) || (reconnect_loc == tree[i]->get_parent()))
				reconnect_loc=tree[i]->get_sibling();
			add_to_branch_list(extra_brn, tree[i]);
			add_to_branch_list(extra_brn, tree[i]->get_parent());
			
			old_sib=tree[i]->get_sibling();
			old_sib->null_parent();
			old_sib->null_sibling();
			root=old_sib;
		}
		tips=tips->next;
	}


	
	extra_brn=start_extra_brn;
	build_tree(subtree_root, subtree_root, left, right);
	old_sib=left;
	if (old_sib->get_brnlen() == 0)
		old_sib->set_brnlen(subtree_root->get_sibling()->get_brnlen()*split_fac);
	old_par=extra_brn->element;
	old_par->set_tip(FALSE);
	old_par->set_taxa_id(-1);
	
	if (reconnect_loc->get_brnlen() == 0) {
		reconnect_loc->set_brnlen(old_par->get_brnlen());
		old_par->set_brnlen(0);
		if (reconnect_loc->get_parent() == 0)
			if (reconnect_loc->get_child(0)->get_brnlen() == 0)
				orig_null=reconnect_loc->get_child(0);
	}
	
	

	reconnect_sib=reconnect_loc->get_sibling();
	reconnect_par=reconnect_loc->get_parent();
	if (reconnect_par != 0) {
		if (reconnect_par->get_child(0) == reconnect_loc)
			child_num=0;
		else
			child_num=1;
	}


	set_as_siblings(reconnect_loc, old_sib);
	reconnect_loc->set_brnlen(reconnect_loc->get_brnlen()*split_fac);
	old_par->set_brnlen(reconnect_loc->get_brnlen()*(1.0-split_fac));
	set_as_parent_child(old_par, reconnect_loc, 0);
	set_as_parent_child(old_par, old_sib, 1);
	if (orig_null != 0)
	{
		orig_null->set_brnlen(old_par->get_child(0)->get_brnlen());
		old_par->get_child(0)->set_brnlen(0);
	}
	
	if (reconnect_sib != 0)
		set_as_siblings(reconnect_sib, old_par);
	if (reconnect_par != 0)
		set_as_parent_child(reconnect_par, old_par, child_num);
	else {
		old_par->null_sibling();
		old_par->null_parent();
	}

	set_root(reconnect_loc);
	diganose_tree(root);
	diganose_tree(root);

}


void Tree::prune_from_tree(Branch *prune_branch)
{
  int sib_num;
  Branch *new_parent;
 
  if (prune_root==0)
    {
      if (prune_branch==find_root() || prune_branch==find_null_branch_id())
	cerr<<"Cannot prune the root or null branch from the tree\n";
      else if (prune_branch!=find_null_branch_id()->get_sibling())
	{
	  prune_root=prune_branch->get_parent();
	  
	  new_parent=prune_root->get_parent();

	  if (new_parent->get_child(0)==prune_root)
	    sib_num=0;
	  else
	    sib_num=1;
	  
	  set_as_parent_child(new_parent, prune_branch->get_sibling(), sib_num);
	  set_as_siblings(new_parent->get_child(0), new_parent->get_child(1));
	  prune_branch->null_sibling();
	  if(prune_root->get_child(0)==prune_branch)
	    prune_root->null_child(1);
	  else
	    prune_root->null_child(0);
	}

      else
	{
	  
	  prune_root=find_null_branch_id();
	  set_as_parent_child(find_root(), find_null_branch_id()->get_child(0), 0);
	  set_as_parent_child(find_root(), find_root()->get_child(0)->get_sibling(), 1);
	  

	  set_as_parent_child(prune_root, prune_branch, 0);
	  prune_root->null_child(1);
	  prune_root->null_parent();
	  prune_branch->null_sibling();
	}
     
    }
  else
    cerr<<"An existing part of the tree has already been pruned--reattach before\ncontinuing\n";
}  //End Tree::prune_from_tree



void Tree::add_back_prune(Branch *add_to)
{  int child_num, prune_child;

  if (prune_root==0)
    cerr<<"Must prune before adding pruned piece back to tree\n";
  else
    {
      if (add_to==find_root())
	cerr<<"Can't add pruned subtree to the root of current tree\n";
      else
	{
	  if (add_to->get_parent()->get_child(0)==add_to)
	    child_num=0;
	  else
	    child_num=1;

	  set_as_parent_child(add_to->get_parent(), prune_root, child_num);
	  set_as_siblings(add_to->get_parent()->get_child(0), add_to->get_parent()->get_child(1));

	  if (prune_root->get_child(0)==0)
	    prune_child=0;
	  else
	    prune_child=1;

	  set_as_parent_child(prune_root, add_to, prune_child);
	  set_as_siblings(prune_root->get_child(0), prune_root->get_child(1));
	  prune_root=0;
	  set_root(tree[0]);
	}
    }

}  //End Tree::add_back_prune




void Tree::diganose_tree(Branch *start)
{
	BOOL error=FALSE; 
  if (start->get_sibling()==0 && start!=root)
    cout<<"Taxa "<<start->get_taxa_id()<<" sibling is null\n";
  if (start->get_parent()==0 && start!=root)
    cout<<"Taxa "<<start->get_taxa_id()<<" parent is null\n";
  if (start->get_child(0)==0 && start->is_tip()==FALSE) 
    cout<<"Taxa "<<start->get_taxa_id()<<" child 0 is null\n";
  if (start->get_child(1)==0 && start->is_tip()==FALSE)
    cout<<"Taxa "<<start->get_taxa_id()<<" child 1 is null\n";
  if (((start->get_child(0) != 0) || (start->get_child(1) != 0)) && (start->is_tip() ==TRUE)) {
    cout<<"Taxa "<<start->get_taxa_id()<<" is tip with child\n";
	error=TRUE;
  }
  if((start->is_tip() == TRUE)  && (strcmp(start->get_name(), "NONE") == 0))
  {
	cout<<"Tip taxa "<<start->get_taxa_id()<<" has invalid name\n";
	error=TRUE;
  }
	//if (((start != root) && (start != find_null_branch()) && start->get_brnlen() == 0))
	//	cerr<<"Zero-length branch "<<start<<" is not root or null\n";
  
  if ((start->is_tip()==FALSE) && (error == FALSE))
    {
      diganose_tree(start->get_child(0));
      diganose_tree(start->get_child(1));
    }
  else if (error == TRUE)
	cout<<"Exiting\n";
}


BOOL Tree::is_root(Branch *the_brn)
{
  if (the_brn->get_parent()==0)
    return(TRUE);
  else
    return(FALSE);
}




Tree::~Tree()
{
	int i;
	Branch_list *temp;
	
	while (start_extra_brn != 0)
	{
		temp=start_extra_brn->next;
		delete start_extra_brn;
		start_extra_brn=temp;
	}

	while (start_tips != 0)
	{
		temp=start_tips->next;
		delete start_tips;
		start_tips=temp;
	}

  if (tree_is_local_mem==TRUE) {
    for (i=0; i<curr_exchange->get_num_branches(); i++)
      delete tree[i];
    
    delete[] tree;
  }
}
//End of Public Functions





void Tree::set_as_siblings(Branch *sib1, Branch *sib2)
{
  sib1->set_sibling(sib2);
  sib2->set_sibling(sib1);
}  //end Tree::set_as_siblings




void Tree::set_as_parent_child(Branch *parent, Branch *child, int child_num)
{
  parent->set_child(child, child_num);
  child->set_parent(parent);
} //End Tree::set_as_parent_child


int Tree::other_child(int child_num)
{
  if (child_num==0)
    return(1);
  else
    return(0);
}


//Private Functions:
void Tree::get_tips(Branch *start, Branch *myroot)
{

	//First progress to the left tip
	if (start->is_tip() == FALSE)
		get_tips(start->get_child(0), myroot);
	else 
		add_to_branch_list(tips, start);	
	
 
	//Check if this branch is a left child (root check performed to avoid
	//seg fault if check parent of root)
	if ((start != myroot) && (start->get_parent()->get_child(0)==start))
    {
	  //If sib is tip, number it and check if it is the target
      if (start->get_sibling()->is_tip() == TRUE)
		add_to_branch_list(tips, start->get_sibling());	
      //If our sib is not a tip, number that subtree
      else {
		get_tips(start->get_sibling(), myroot);
		return;
	  }
    }
  //If we are at a right child--just return
  else
    return;
}



void Tree::build_tree(Branch *tree_loc, Branch *myroot, Branch *&child1, Branch *&child2)
{
	Branch *new_brn, *new_sib;

	//First progress to the left tip
	if (tree_loc->is_tip() == FALSE) {
		build_tree(tree_loc->get_child(0), myroot, child1, child2);
		new_brn=extra_brn->element;
		extra_brn=extra_brn->next;
		(*new_brn)=(*tree_loc);
		set_as_parent_child(new_brn, child1, 0);
		set_as_parent_child(new_brn, child2, 1);
	}	
	else {
		new_brn=extra_brn->element;
		
		extra_brn=extra_brn->next;
		(*new_brn)=(*tree_loc);	
	}
 
	//Check if this branch is a left child (root check performed to avoid
	//seg fault if check parent of root)
	if ((tree_loc != myroot) && (tree_loc->get_parent()->get_child(0)==tree_loc))
    {
		if (tree_loc->get_sibling()->is_tip() == TRUE)
		{
			new_sib=extra_brn->element;
			extra_brn=extra_brn->next;
			(*new_sib)=(*tree_loc->get_sibling());		  
		}
		//If our sib is not a tip, traverse that subtree
		else {
			build_tree(tree_loc->get_sibling()->get_child(0), myroot, child1, child2);
			new_sib=extra_brn->element;
			extra_brn=extra_brn->next;
			(*new_sib)=(*tree_loc->get_sibling());
			set_as_parent_child(new_sib, child1, 0);
			set_as_parent_child(new_sib, child2, 1);
		}
		set_as_siblings(new_brn, new_sib);
		child1=new_brn;
		child2=new_sib;
		return;
    }
	else if (tree_loc == myroot) {
		child1=new_brn;
		return;
	}
  //If we are at a right child--just return
	else
		return;	
}



Branch* Tree::invert_clade(Branch *new_parent, Branch *old_parent, Branch *old_sibling)
{
 Branch *old_par, *old_sib;

 old_par=old_parent->get_parent();
 old_sib=old_parent->get_sibling();

 set_as_parent_child(new_parent, old_parent, 0);
 set_as_parent_child(new_parent, old_sibling, 1);
 set_as_siblings(old_parent, old_sibling);


 if (old_par!=0)
     invert_clade(old_parent, old_par, old_sib);

 else
   return(old_parent);

} //End Tree::invert_clade





Tree * Tree::split_tree(Branch *split_point)
  //This function takes a rooted tree, splits it at branch 
  //split_point and returns a pointer to one of the two new trees
  //(The other remains in the original tree pointer
{
  int old_par_child_num;
  Tree *new_tree;
  Branch *old_par, *old_sib, *old_par_sib;

  if (split_point != root) {
    old_par=split_point->get_parent();
    old_sib=split_point->get_sibling();
  
    if (old_par != root) { 
      old_par_sib=old_par->get_sibling();
      if (old_par->get_parent()->get_child(0) == old_par)
	old_par_child_num=0;
      else
	old_par_child_num=1;
      
      set_as_parent_child(old_par->get_parent(), old_sib, old_par_child_num);
      set_as_siblings(old_sib, old_par_sib);
      extra_branch = old_par;
      set_root(old_sib);
   
    }
    else 
      {
	old_sib->null_parent();
	old_sib->null_sibling();
	extra_branch = root;
	set_root(old_sib);
      }
    
    split_point->null_parent();
    split_point->null_sibling();
    new_tree=new Tree(curr_exchange, split_point);

    return(new_tree);
  }
  else
    return(0);

}



Branch * Tree::join_trees(Tree *tree1, Tree *tree2, Branch *join1, Branch *join2, int choosetree)
  //Takes pointers to two trees, with a branch on each to join the trees at.
  //Reroots the non-base tree at its join point, and then joins them
{
  int old_par_child_num;
  double brn_swap;
  Branch *tree_root, *add_to, *old_par, *old_sib, *old_join_null=0, *old_join_root=0;
  Tree *basetree, *jointree;


  //Descide which tree is the base
  if (choosetree == 0) {
    basetree=tree1;
    jointree=tree2;
    tree_root=join2;
    add_to=join1;
  }  
  else {
    basetree=tree2;
    jointree=tree1;
    tree_root=join1;
    add_to=join2;
  }

  
  if (jointree->find_null_branch() != 0) 
	old_join_null=jointree->find_null_branch();
  old_join_root=jointree->find_root();
  
  //Re-roots the non-base tree at it's join point
  if ((tree_root != jointree->find_root()) && (tree_root != jointree->find_null_branch_id()))
      jointree->re_root_tree(tree_root);

 
  //Finds the "real" root, whose child is tree_root
  if (jointree->find_root() != tree_root)
    tree_root=tree_root->get_parent();

  //Joins the two trees
  if (add_to != basetree->find_root())
    {
      old_sib=add_to->get_sibling();
      old_par=add_to->get_parent();
      
      if (old_par->get_child(0) == add_to)
	old_par_child_num=0;
      else
	old_par_child_num=1;
     
      set_as_siblings(add_to, tree_root);
      set_as_parent_child(extra_branch, add_to, 0);
      set_as_parent_child(extra_branch, tree_root, 1);
      set_as_siblings(extra_branch, old_sib);
      set_as_parent_child(old_par, extra_branch, old_par_child_num);
     
	  if (add_to->get_brnlen() == 0) {
		add_to->set_brnlen(extra_branch->get_brnlen());
		extra_branch->set_brnlen(0);
	  }

      basetree->set_root(add_to);
    
	}
  else {
    extra_branch->null_parent();
    extra_branch->null_sibling();
   

    set_as_parent_child(extra_branch, add_to, 0);
    set_as_parent_child(extra_branch, tree_root, 1);
    set_as_siblings(add_to, tree_root);
	if (add_to->get_child(0) != 0)
		if (add_to->get_child(0)->get_brnlen() == 0)
		{
			add_to->get_child(0)->set_brnlen(extra_branch->get_brnlen());
			extra_branch->set_brnlen(0);
		}
	
    basetree->set_root(add_to);
  }
	if (old_join_root->get_brnlen()==0)
	{
		old_join_root->set_brnlen(basetree->find_root()->get_brnlen());
		basetree->find_root()->set_brnlen(0);
	}
	if (old_join_null != 0)
		if (old_join_null->get_brnlen() == 0) {
			old_join_null->set_brnlen(basetree->find_null_branch()->get_brnlen());
			basetree->find_null_branch()->set_brnlen(0);
		}
	if((extra_branch->get_brnlen() == 0) &&(extra_branch != basetree->find_root()) && (extra_branch != basetree->find_null_branch()))
	{
		extra_branch->set_brnlen(basetree->find_null_branch()->get_brnlen());
		basetree->find_null_branch()->set_brnlen(0);
	}
	
  root=basetree->find_root();
  left_tip=basetree->get_leftmost_tip();

  diganose_tree(basetree->find_root());
  if ((basetree->find_root()->get_sibling() !=0 ) || (basetree->find_root()->get_parent() != 0))
    cerr<<"Error joining tree--root is not bottom: sib: "<<basetree->find_root()->get_sibling()<<" Parent: "
	<<basetree->find_root()->get_parent()<<"\n";
  return(add_to);
}



void Tree::set_num_nonsyn_params()
{
	int i, j;
	BOOL used;


	max_p_nonsyn=0;
	max_evol_pattern=new int [curr_exchange->get_num_nonsyn_params()];
	for(i=0; i<curr_exchange->get_num_nonsyn_params(); i++)
		max_evol_pattern[i]=0;


	switch(curr_exchange->get_model()) {
	case MG_94_JC:
	case MG_94_K2P:
	case MG_94_HKY:
		for(i=0; i<curr_exchange->get_num_branches(); i++) 
			if(tree[i]->get_p_nonsyn_num()+1>max_p_nonsyn) {
				max_p_nonsyn=tree[i]->get_p_nonsyn_num()+1;
				max_evol_pattern[0]=tree[i]->get_p_nonsyn_num();
			}
		
		break;
	case C_00_JC:
	case C_00_K2P:
	case C_00_HKY:
		for(i=0; i<curr_exchange->get_num_branches(); i++) {
			if(tree[i]->get_p_nonsyn_num()>max_evol_pattern[0])
				max_evol_pattern[0]=tree[i]->get_p_nonsyn_num();
			
			if(tree[i]->get_p_intergroup_num()>max_evol_pattern[1])
				max_evol_pattern[1]=tree[i]->get_p_intergroup_num();


			if(tree[i]->get_p_nonsyn_num()+1>max_p_nonsyn)
				max_p_nonsyn=tree[i]->get_p_nonsyn_num()+1;
			if(tree[i]->get_p_intergroup_num()+1>max_p_nonsyn)
				max_p_nonsyn=tree[i]->get_p_intergroup_num()+1;
		}
		break;
	case LCAP_JC:
	case LCAP_K2P:
	case LCAP_HKY:
			for(i=0; i<curr_exchange->get_num_branches(); i++) {
				for(j=0; j<curr_exchange->get_num_aa_props()+1; j++) {
					if(tree[i]->get_aa_prop_num(j)>max_evol_pattern[j])
						max_evol_pattern[j]=tree[i]->get_aa_prop_num(j);

					if(tree[i]->get_aa_prop_num(j) +1 > max_p_nonsyn)
						max_p_nonsyn=tree[i]->get_aa_prop_num(j) +1;
				}
			}
		break;	
	case AAMULMAT_JC:
	case AAMULMAT_K2P:
	case AAMULMAT_HKY:
		for(i=0; i<curr_exchange->get_num_branches(); i++) {
				for(j=0; j<curr_exchange->get_num_matrices(); j++) {
					if(tree[i]->get_matrix_coeff_num(j)>max_evol_pattern[j])
						max_evol_pattern[j]=tree[i]->get_aa_prop_num(j);

					if(tree[i]->get_matrix_coeff_num(j) +1 > max_p_nonsyn)
						max_p_nonsyn=tree[i]->get_matrix_coeff_num(j) +1;
				}
			}
			
		break;
	}


	get_num_distinct_evol_patterns();
	cout<<"Identified "<<num_patterns<<" distinct patterns of non-synonymous substitutions\n";
    curr_exchange->set_num_p_nonsyn(max_p_nonsyn, num_patterns);



	for(i=0; i<curr_exchange->get_num_branches(); i++)
		cout<<"Branch "<<i<<" Name: "<<tree[i]->get_name()<<" PATERN: "<<
		tree[i]->get_nonsyn_pattern()<<" PNS#: "<<tree[i]->get_p_nonsyn_num()
		<<" PIG:#: "<<tree[i]->get_p_intergroup_num()<<" AA0#:"<<tree[i]->get_aa_prop_num(0)<<endl;

	switch (curr_exchange->get_model()) {
	case C_00_JC:
	case C_00_K2P:
	case C_00_HKY:
		for(i=0; i<max_p_nonsyn; i++)
		{
			if (max_evol_pattern[0] >= i)
				used=TRUE;
			else
				used=FALSE;

			curr_exchange->set_p_non_syn_used(i, used);
			
			if (max_evol_pattern[1] >= i)
				used=TRUE;
			else
				used=FALSE;

			curr_exchange->set_p_inter_group_used(i, used);
		
		}
		break;
	case LCAP_JC:
	case LCAP_K2P:
	case LCAP_HKY:
		for(i=0; i<max_p_nonsyn; i++)
		{
			for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++) {
				if (max_evol_pattern[j] >= i)
					used=TRUE;
				else
					used=FALSE;

				curr_exchange->set_aa_property_used(i, j, used);
			}
		
		}
		break;
	case AAMULMAT_JC:
	case AAMULMAT_K2P:
	case AAMULMAT_HKY:
		for(i=0; i<max_p_nonsyn; i++)
		{
			for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++) {
				if (max_evol_pattern[j] >= i)
					used=TRUE;
				else
					used=FALSE;

				curr_exchange->set_matrix_coeff_used(i, j, used);
			}
		
		}	
		break;



	}

	delete[] max_evol_pattern;

}


void Tree::get_num_distinct_evol_patterns()
{
	int i, j, *patt, nonsyn_match;
	BOOL matches, same;
	EPattern_list *listpoint;

	num_patterns=1;
	the_patterns=0;

	if(curr_exchange->get_num_nonsyn_params() == 1) {
		num_patterns=max_p_nonsyn;
	}
	else {
		patt=new int[curr_exchange->get_num_nonsyn_params()];

		for(i=0; i<curr_exchange->get_num_branches(); i++) {
			switch(	curr_exchange->get_model()) {
			case C_00_JC:
			case C_00_K2P:
			case C_00_HKY:
				patt[0]=tree[i]->get_p_nonsyn_num();
				patt[1]=tree[i]->get_p_intergroup_num();
				break;
			case LCAP_JC:
			case LCAP_K2P:
			case LCAP_HKY:
				for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++)
					patt[j]=tree[i]->get_aa_prop_num(j);
				break;
			case AAMULMAT_JC:
			case AAMULMAT_K2P:
			case AAMULMAT_HKY:
				for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++)
					patt[j]=tree[i]->get_matrix_coeff_num(j);
				break;
				
				break;

			}
			if(the_patterns != 0) {
				listpoint=the_patterns;

				the_patterns=start_patt;

				matches=FALSE;
				while(the_patterns != 0) {
					same=TRUE;
					for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++) {
						if (the_patterns->pattern[j] != patt[j])
							same=FALSE;
					}
					if (same == TRUE) {
						matches = TRUE;
						nonsyn_match=the_patterns->pattern_branch;
					}
				
					the_patterns=the_patterns->next;
				}

				the_patterns=listpoint;
				if (matches == FALSE) {
					num_patterns++;
					tree[i]->set_nonsyn_pattern(num_patterns-1);
					the_patterns->next=new EPattern_list;
					the_patterns->next->last=the_patterns;
					the_patterns=the_patterns->next;
					the_patterns->next=0;
					the_patterns->pattern=new int [curr_exchange->get_num_nonsyn_params()];
					the_patterns->pattern_branch=i;
					for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++)
						the_patterns->pattern[j]=patt[j];

				}
				else
					tree[i]->set_nonsyn_pattern(tree[nonsyn_match]->get_nonsyn_pattern());
			}
			else {
				the_patterns=new EPattern_list;
				the_patterns->pattern=new int[curr_exchange->get_num_nonsyn_params()];
				the_patterns->pattern_branch=i;
				start_patt=the_patterns;
				the_patterns->last=the_patterns->next=0;
				for(j=0; j<curr_exchange->get_num_nonsyn_params(); j++)
					the_patterns->pattern[j]=patt[j];
				tree[i]->set_nonsyn_pattern(num_patterns-1);
			}

		}

		the_patterns=start_patt;

		while(the_patterns != 0)
		{
			delete[] the_patterns->pattern;
			listpoint=the_patterns->next;
			delete the_patterns;
			the_patterns=listpoint;
		}

	}


}



void add_to_branch_list(Branch_list *&list, Branch *new_ele)
{
	list->element=new_ele;
	if (list->next==0) {
		list->next=new Branch_list;
		list=list->next;
		list->next=0;
	}
	else
		list=list->next;
	list->element=0;		
}
